==> snapstraction-0.8.1-py2-none-any.whl: snapstraction/topology.py <==
"""Utilities for gathering topologies from a mesh"""
from __future__ import print_function

import binascii
import logging

from collections import deque, Iterable, defaultdict

from tornado.gen import coroutine, Return

from .gus import NoResponseError

from .node import get_vmstat
from .node_cache import NodeCache
from .util import fix_up_target

log = logging.getLogger(__name__)


def _chunk_string(string, size):
    """Break a string into fixed-size chunks"""

    if not string:
        return []

    if not size:
        size = len(string)

    return [string[i:i + size] for i in range(0, len(string), size)]


def _parse_topo_response(src, response):
    """
    Parse the Topology (vmStat 12) response chunk.

    Returns a dict of Edges (or {} if the response could not be processed):
    * A, B: lq A->B (i.e. "how well B could hear A")
    * B, A: lq B->A (i.e. "how well A could hear B")
    """

    if not src:
        return {}

    if not response:
        return {}

    if len(response) != 5:
        return {}

    dst = binascii.hexlify(response[0:3])

    lq_src_to_dst = ord(response[3])
    lq_dst_to_src = ord(response[4])

    return {(src, dst): lq_src_to_dst,
            (dst, src): lq_dst_to_src}


def _mean(l):
    if isinstance(l, Iterable):
        return int(sum(l) / len(l))
    else:
        # If it's somehow not a list-ish thing, just return it.
        # We probably got called multiple times on the same topo.
        # (Mean of single value is that value.)
        return l


def _merge_topo_responses(responses):
    """Merge multiple node responses, averaging multiple lqs"""

    if not responses:
        return {}

    return {(src, dst): _mean(lqs) for (src, dst), lqs in responses.iteritems()}


@coroutine
@fix_up_target
def take_node_census(gus, target):
    """Take a census of a single node"""
    data = {}

    try:
        val = yield get_vmstat(gus, target, 12)
    except NoResponseError:
        raise Return({'root': target, 'topology': {}})

    parsed = (_parse_topo_response(target, s) for s in _chunk_string(val['data'], 5))

    with NodeCache() as cache:
        for response in parsed:
            for (src, dst), lq in response.iteritems():
                data[(src, dst)] = lq

                cache.add(src)
                cache.add(dst)

    raise Return({'root': target, 'topology': data})


@coroutine
@fix_up_target
def take_network_census(gus, target):
    """
    Take a census of an entire network, starting with a single target

    Breadth-first from the starting point.
    """
    data = defaultdict(list)

    nodes = deque()
    polled = set()
    nodes.append(target)

    while True:
        try:
            n = nodes.pop()
            while n in polled:
                n = nodes.pop()
        except IndexError:
            break

        try:
            log.debug("Getting Node Topology for: %r", n)
            polled.add(n)
            node_topo = yield take_node_census(gus, n)
            log.debug("Got Node Topology: %r", node_topo)

            for (src, dst), lq in node_topo['topology'].iteritems():
                data[(src, dst)].append(lq)

                if src not in polled:
                    nodes.append(src)
                if dst not in polled:
                    nodes.append(dst)

        except:
            log.exception("Gathering topology for node %s failed", n)
            continue

    merged = _merge_topo_responses(data)

    raise Return({'root': target, 'topology': merged})

==> snapstraction-0.9.1-py2-none-any.whl: snapstraction/topology.py <==
"""Utilities for gathering topologies from a mesh"""
from __future__ import print_function

import binascii
import logging

from collections import deque, Iterable, defaultdict

from tornado.gen import coroutine, Return, sleep

from snappy import SnapConstantsHelper as SnapConst

from .gus import NoResponseError

from .node_cache import NodeCache
from .util import fix_up_target

log = logging.getLogger(__name__)


def _chunk_string(string, size):
    """Break a string into fixed-size chunks"""

    if not string:
        return []

    if not size:
        size = len(string)

    return [string[i:i + size] for i in range(0, len(string), size)]


def _parse_topo_response(src, response):
    """
    Parse the Topology (vmStat 12) response chunk.

    Returns a dict of Edges (or {} if the response could not be processed):
    * A, B: lq A->B (i.e. "how well B could hear A")
    * B, A: lq B->A (i.e. "how well A could hear B")
    """

    if not src:
        return {}

    if not response:
        return {}

    if len(response) != 5:
        return {}

    dst = binascii.hexlify(response[0:3])

    lq_src_to_dst = ord(response[3])
    lq_dst_to_src = ord(response[4])

    return {(src, dst): lq_src_to_dst,
            (dst, src): lq_dst_to_src}


def _mean(l):
    if isinstance(l, Iterable):
        return int(sum(l) / len(l))
    else:
        # If it's somehow not a list-ish thing, just return it.
        # We probably got called multiple times on the same topo.
        # (Mean of single value is that value.)
        return l


def _merge_topo_responses(responses):
    """Merge multiple node responses, averaging multiple lqs"""

    if not responses:
        return {}

    return {(src, dst): _mean(lqs) for (src, dst), lqs in responses.iteritems()}


@coroutine
@fix_up_target
def take_node_census(gus, target):
    """Take a census of a single node"""
    data = {}

    class nonlocal:
        pass

    nonlocal.time_remaining = gus.network.default_cb_timeout

    # Set up response handler
    def on_vmStat(src, *args):
        """tellVmStat callback"""
        log.debug("tellVmStat: %s %r", src, args)
        if src != target:
            return

        (word, d) = args

        word = word & 0xFFFF

        code = word & 0xFF

        # Make sure it's our tellVmStat...
        if code != SnapConst.VM_TOPOLOGY:
            return

        parsed = (_parse_topo_response(target, s) for s in _chunk_string(d, 5))
        for response in parsed:
            log.debug("Response: %r", response)
            for (src, dst), lq in response.iteritems():
                data[(src, dst)] = lq

        # Refresh our timeout
        nonlocal.time_remaining = gus.network.default_cb_timeout

    # Register our tellVmStat handler
    gus.on_rpc('tellVmStat', on_vmStat)

    # Initial send
    for _ in xrange(gus.network.default_cb_retries):
        try:
            yield gus.ucast_rpc(target, 'vmStat', SnapConst.VM_TOPOLOGY)
            yield sleep(gus.network.default_cb_timeout)
            if data:
                break
        except NoResponseError:
            raise Return({'root': target, 'topology': {}})

    # Wait for all remaining replies
    # Receiving a response refreshes the timeout
    while nonlocal.time_remaining > 0:
        nonlocal.time_remaining -= 0.1
        yield sleep(0.1)

    gus.off_rpc('tellVmStat', on_vmStat)

    # Add nodes to the cache
    with NodeCache() as cache:
        for (src, dst) in data.keys():
            cache.add(src)
            cache.add(dst)

    raise Return({'root': target, 'topology': data})


@coroutine
@fix_up_target
def take_network_census(gus, target, max_lq=128):
    """
    Take a census of an entire network, starting with a single target

    Breadth-first from the starting point.
    """
    data = defaultdict(list)

    nodes = deque()
    polled = set()
    nodes.append(target)

    while True:
        try:
            n = nodes.pop()
            while n in polled:
                n = nodes.pop()
        except IndexError:
            break

        try:
            log.debug("Getting Node Topology for: %r", n)
            polled.add(n)
            node_topo = yield take_node_census(gus, n)
            log.debug("Got Node Topology: %r", node_topo)

            for (src, dst), lq in node_topo['topology'].iteritems():
                data[(src, dst)].append(lq)

                # Only add to the queue to poll if the lq is better than some value
                if src not in polled and lq < max_lq:
                    nodes.append(src)
                if dst not in polled and lq < max_lq:
                    nodes.append(dst)

        except:
            log.exception("Gathering topology for node %s failed", n)
            continue

    merged = _merge_topo_responses(data)

    raise Return({'root': target, 'topology': merged})
